/************************************************************************
* Sourcefile Seriell-I/O Drucklogger
*
* Version:	1.0
* Compiler:	GCC 4.1.1 (arm-elf-gcc)
* Datum:	23.01.2007 - 29.01.2007
* Autor:	Hermann Gnther, Saxotec GmbH & Co.KG
************************************************************************/

#include "Utils/LPC214x.h" 
#include "Utils/type.h"
#include "Utils/irq.h"
#include "LPCUSB/usbapi.h"

#include "serio.h"
#include "config.h"
#include "console.h"

static void USBHandler( void ) __irq;

#define MAX_BYTE_COUNT	400


#define BULK_IN_EP		0x82
#define BULK_OUT_EP		0x05

#define MAX_PACKET_SIZE	64

#define LE_WORD(x)		((x)&0xFF),((x)>>8)


U8 rs_buffer[MAX_BYTE_COUNT];
U8 ack;
BOOL new_command=FALSE;
U8 current_interface=SERIO_USB;
//Paritt
U8 parity;
//Lnge und Zhler USB Senden
U8 * s_ptr;
int s_length=0;
int s_cnt=0;



static U8 abVendorReqData[1];
static const U8 abDescriptors[] = {

/* Device descriptor */
	0x12,              		
	DESC_DEVICE,       		
	LE_WORD(0x0200),		// bcdUSB	
	0xFF,              		// bDeviceClass
	0x00,              		// bDeviceSubClass
	0x00,              		// bDeviceProtocol
	MAX_PACKET_SIZE0,  		// bMaxPacketSize
	LE_WORD(USB_ID_VENDOR),	// idVendor
	LE_WORD(USB_ID_PRODUCT),// idProduct
	LE_WORD(USB_RELEASE),	// bcdDevice
	0x01,              		// iManufacturer
	0x02,              		// iProduct
	0x03,              		// iSerialNumber
	0x01,              		// bNumConfigurations

// configuration
	0x09,
	DESC_CONFIGURATION,
	LE_WORD(0x20),  		// wTotalLength
	0x01,  					// bNumInterfaces
	0x01,  					// bConfigurationValue
	0x00,  					// iConfiguration
	0x80,  					// bmAttributes
	0x32,  					// bMaxPower

// interface
	0x09,   				
	DESC_INTERFACE, 
	0x00,  		 			// bInterfaceNumber
	0x00,   				// bAlternateSetting
	0x02,   				// bNumEndPoints
	0xFF,   				// bInterfaceClass
	0x00,   				// bInterfaceSubClass
	0x00,   				// bInterfaceProtocol
	0x00,   				// iInterface

// bulk in
	0x07,   		
	DESC_ENDPOINT,   		
	BULK_IN_EP,				// bEndpointAddress
	0x02,   				// bmAttributes = BULK
	LE_WORD(MAX_PACKET_SIZE),// wMaxPacketSize
	0,						// bInterval   		

// bulk out
	0x07,   		
	DESC_ENDPOINT,   		
	BULK_OUT_EP,			// bEndpointAddress
	0x02,   				// bmAttributes = BULK
	LE_WORD(MAX_PACKET_SIZE),// wMaxPacketSize
	0,						// bInterval   		

// string descriptors
	0x04,
	DESC_STRING,
	LE_WORD(0x0409),

	// manufacturer string
	0x02 + USB_MANUFACTURER_LENGTH,
	DESC_STRING,
	USB_MANUFACTURER,

	// product string
	0x02 + USB_PRODUCT_LENGTH,
	DESC_STRING,
	USB_PRODUCT,

	// serial number string
	0x12 + USB_HWV_LENGTH + USB_SWV_LENGTH,
	DESC_STRING,
	'H', 0, 'W', 0, ' ', 0, USB_HWV, ',', 0, ' ', 0, 'S', 0, 'W', 0, ' ', 0, USB_SWV,
	
	// terminator
	0
};



static void _HandleBulkIn(U8 bEP, U8 bEPStatus)
{
	int iChunk;
	
	iChunk = MIN(MAX_PACKET_SIZE, s_length);
putchar('-');
putchar('0'+iChunk/10);
putchar('0'+iChunk%10);
putchar('-');
	if (iChunk == 0) {
		return;
	}
	
	// send next part
	USBHwEPWrite(bEP, s_ptr, iChunk);
	
	s_ptr += iChunk;
	s_length -= iChunk;

}


static void _HandleBulkOut(U8 bEP, U8 bEPStatus)
{
	int iChunk;
	U8 r_buf[MAX_PACKET_SIZE];		//Empfangspuffer (mehr als MAX_PACKET_SIZE auf einmal lesen geht sowieso nicht)
	U8 * r_ptr=r_buf;				//Zeiger auf Empfangspuffer
	static BOOL receiving=FALSE;
	static U8 replace_next=0;
	static U8 * grs_ptr;			//Zeiger auf Befehlspuffer
	
	// get next part
	iChunk = USBHwEPRead(bEP, r_buf, MAX_PACKET_SIZE);

	while(iChunk!=0){
		if(!receiving){
			if(*r_ptr==SOH){		//Start eines Paketes
				receiving=TRUE;
				grs_ptr=rs_buffer;	//Zeiger auf Anfang von Puffer
				parity=SOH;
			}
		}
		else{
			if(*r_ptr!=EOT){
				if(*r_ptr==DLE){	//Ersetztes Steuerzeichen ...
					replace_next=1;
					continue;
				}
				if(replace_next){	//... rekonstruieren
					*r_ptr -= DLE;
					replace_next=0;
				}
				parity ^= *r_ptr;	//Paritt berechnen
			}
			else{
				if(parity)			//Paritt stimmt nicht
					ack=0;
				else
					ack=1;
				receiving=FALSE;
				new_command=TRUE;
				grs_ptr--;			//Paritt mit EOT berschreiben	
			}
			*grs_ptr=*r_ptr;		//Zeichen in Puffer schreiben
			grs_ptr++;
		}
		r_ptr++;
		iChunk--;
	}
}


/*************************************************************************
	HandleVendorRequest
	===================
		Handles vendor specific requests
		
	Control transfer fields:
	* request:	0x01 = prepare memory read
				0x02 = prepare memory write
	* index:	ignored
	* value:	ignored
	* data:		ignored
		
**************************************************************************/
static BOOL HandleVendorRequest(TSetupPacket *pSetup, int *piLen, U8 **ppbData)
{
	switch (pSetup->bRequest) {
	
	// prepare read
	case 0x01:
		// send initial packet
		_HandleBulkIn(BULK_IN_EP, 0);
		*piLen = 0;
		break;
		
	// prepare write
	case 0x02:
		*piLen = 0;
		break;

	default:
		return FALSE;
	}
	return TRUE;
}

static void USBHandler (void) __irq 
{
	IENABLE;				//Falls nested interrupts eingestellt (siehe Utils/irq.h), Interrupts aktivieren
	
//Call USB interrupt handler
	USBHwISR();
	
    IDISABLE;				//Falls nested interrupts eingestellt (siehe Utils/irq.h), Interrupts deaktivieren
    VICVectAddr = 0;		/* Acknowledge Interrupt */
}


void serio_select_interface(U8 interface){
	current_interface=interface;
}


void serio_init(U8 interface){

	if(interface==SERIO_USB){
		// initialise USB stack
		USBInit();
		// register device descriptors
		USBRegisterDescriptors(abDescriptors);
		// override standard request handler
		USBRegisterRequestHandler(REQTYPE_TYPE_VENDOR, HandleVendorRequest, abVendorReqData);
		// register endpoints
		USBHwRegisterEPIntHandler(BULK_IN_EP, _HandleBulkIn);
		USBHwRegisterEPIntHandler(BULK_OUT_EP, _HandleBulkOut);
		
		//USB Interrupt installieren
		install_irq( USB_INT, (void *)USBHandler );
		
		// connect to bus
		USBHwConnect(TRUE);
	}
	else if(interface==SERIO_UART){
		
	}
	serio_select_interface(interface);
}


U8 serio_get_command(U8 ** rec_buffer){
	if(new_command){
		new_command=FALSE;
		*rec_buffer=rs_buffer;	//Pointer auf Empfangspuffer richten
		return ack;
	}
	else
		return 2;
}

void serio_send_start(U8 acknowledge){
	if(current_interface==SERIO_USB){
		rs_buffer[0]=acknowledge;
		s_ptr=rs_buffer;
		if(acknowledge==NAK){
			s_length=1;
			return;
		}
		s_cnt=1;
		parity=ACK;
	}
	else if(current_interface==SERIO_UART){
		
	}
	
}

void serio_send_stop(void){
	if(current_interface==SERIO_USB){
		rs_buffer[s_cnt]=parity;
		s_cnt++;
		rs_buffer[s_cnt]=EOT;
		s_length=s_cnt+1;
	}
	else if(current_interface==SERIO_UART){
		
	}
	
}

void serio_send_char(U8 data){
	if(current_interface==SERIO_USB){
		if(s_cnt==MAX_BYTE_COUNT-2)					//Parity und EOT muss auch noch reinpassen
			return;
    	parity ^= data;
		if(data==0x04 || data==0x06 || data==0x15 || data==0x10){	//EOT,ACK,NAK und DLE ersetzen
			rs_buffer[s_cnt]=0x10;
			s_cnt++;
		    data += 0x10;
		}
		rs_buffer[s_cnt]=data;
		s_cnt++;
	}
	else if(current_interface==SERIO_UART){
		
	}
}

void serio_send_block(U8 * data, U16 number){
	while(number)
	{
		serio_send_char(*data);
		data++;
		number--;
	}
}

void serio_send_string(U8 * String){
	while (*String)
	{
		serio_send_char(*String);
		String++;
	}
}
